// performs nicolas balayage algorithm (general case)

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <time.h>

#include "../library/nrutil.h"
#include "../library/prototypes.h"
#include "../library/2D(k,th)transforms.h"
#include "../library/ProjInfo.h"
#include "../library/PGM.h"


#define CPUTIME (getrusage(RUSAGE_SELF,&ruse), ruse.ru_utime.tv_sec + ruse.ru_stime.tv_sec + 1e-6 * (ruse.ru_utime.tv_usec + ruse.ru_stime.tv_usec))

struct rusage ruse;

extern int getrusage();

typedef struct
{
	int p;
	int q;
	int length;
	int max;
	int min;
	int *ORIGINptr;
	int offset;
	int row_low;
	int row_high;
}projPARAM;

typedef struct
{
	int x;
	int y;
	int proj_no;
	int *PIXELptr;
}reconLIST;

typedef struct
{
	int x;
	int proj_no;
	int done;
}dependLIST;


int GaussNoiseInt(int mean, int range)
{
	double x=1.0, y=1.0;
	while(y > exp(-24.5*x*x))
	{
		x = (double)rand()/(double)RAND_MAX - 0.5;
		y = (double)rand()/(double)RAND_MAX;
	}
	return floor(x*(double)range + 0.5) + mean;
}

void ImageTranspose(int *image,int Q,int P)
{
	int i, j, *buffer, *tempIN, *tempOUT;

	buffer = (int *)calloc(P*Q,sizeof(int));
	ImageCopy(buffer,image,P*Q);
	if(Q>=P)
	{
		tempIN = buffer;
		for(i=0;i<P;i++)
		{	
			tempOUT = image + i;
			for(j=0;j<Q;j++)
			{
				*tempOUT = *tempIN++;
				tempOUT += P;
			}
		}
	}
	else
	{
		tempOUT = image;
		for(i=0;i<Q;i++)
		{	
			tempIN = buffer + i;
			for(j=0;j<P;j++)
			{
				*tempOUT++ = *tempIN;
				tempIN += Q;
			}
		}
	}
	free(buffer);
}


int ReadPROJdataPGM(char *inputname, int *data, int N)
{
/* for reading PGM files of the following format:

P2, P3 or P5
# anything after hash is ignored 
width(= # of proj) height(= 2 for p and q in each row)
greymax(redundant in this case)

*/

/* variables */
	char a, b, c[50];
	int *tmptr, width, height, greymax;
	FILE *fileptr;

	tmptr = data;

	printf("FILE %s ", inputname);
 	if( (fileptr = fopen(inputname, "rb")) == NULL)
	{
		printf("could not be opened.\n");
		exit(0);
	}
	printf("opened OK.\n");

 	fscanf(fileptr, "%c", &a);
 	fscanf(fileptr, "%c", &b);
	fscanf(fileptr, "\n");
	fgets(c, 50, fileptr);

	printf("TYPE: %c%c\nINFO: %s",a,b,c);

	fscanf(fileptr, "%i", &width);
	fscanf(fileptr, "%i", &height);
	fscanf(fileptr, "%i", &greymax);
	fscanf(fileptr, "\n");
	printf("# OF PROJECTIONS:...%d\n",width);

/* check image size is sufficient */
	if((N-(width*height)<0))
	{
		printf("ERROR: not enough memory to read in projection data...\n");
 		exit(0);
	}

/* scan in data*/
	while(!feof(stdin) && N--)
	{
		fscanf(fileptr, "%i", tmptr++ );
	}
	return width;
}

int MinINT(int elem1, int elem2)
{
	if(elem1<=elem2)
		return elem1;
	else
		return elem2;
}

int MaxINT(int elem1, int elem2)
{
	if(elem1>=elem2)
		return elem1;
	else
		return elem2;
}

int CompareP(elem1,elem2)
projPARAM *elem1; projPARAM *elem2;
{
	int a=elem1[0].p,b=elem2[0].p;
	return (a>b) ? 1 : (a<b) ? -1 : 0;
}

int CompareSlope(elem1,elem2)
projPARAM *elem1; projPARAM *elem2;
{
	double a=(double)elem1[0].p/(double)elem1[0].q,b=(double)elem2[0].p/(double)elem2[0].q;
	return (a>b) ? 1 : (a<b) ? -1 : 0;
}

int main(int argc, void *argv[])
{
	int P, Q, Q_n, h, i, j, k, l, x, y, m, n, bin, no_proj, length_proj, temp, temp_row, temp_col;
	int *image, *projection, *projection_copy, *tempIMG, *tempPRJ;
	int **bin_pointer, **BINptr;
	int S_minus=0, S_plus=0, S_total, r, max_row, min_row, offset, start_of_Next, end_of_Next, NoGood, r_max, r_min, q_up_max, q_down_max;
	projPARAM *proj, *PROJptr;
	reconLIST *Recon, *RECONptr;
	dependLIST *Depend;
	double t0, t1;
	time_t u1,u2;
	char filename[50];
		
// read in image width P
	if( sscanf(argv[1],"%d",&P) == 0 )
	{
		printf("ERROR: could not read in image width, P.\n");
		exit(0);
	}

// read in image height Q
	if( sscanf(argv[2],"%d",&Q) == 0 )
	{
		printf("ERROR: could not read in image height, Q.\n");
		exit(0);
	}

// allocate memory for image
	if( (image=(int *)calloc(P*Q,sizeof(int))) == NULL )
	{
		printf("ERROR: Image memory could not be allocated.\n");
		exit(0);
	}
	
// allocate memory for LISTS
	if( (Recon=(reconLIST *)calloc(Q,sizeof(reconLIST))) == NULL )
	{
		printf("ERROR: LIST memory could not be allocated.\n");
		exit(0);
	}
	if( (Depend=(dependLIST *)calloc(P*Q,sizeof(dependLIST))) == NULL )
	{
		printf("ERROR: LIST memory could not be allocated.\n");
		exit(0);
	}

// read in proj data
	no_proj = log2(Q)+1;

	// allocate memory for projection parameters
	if( (proj=(projPARAM *)calloc(no_proj,sizeof(projPARAM))) == NULL )
	{
		printf("ERROR: Projection parameter memory could not be allocated.\n");
		exit(0);
	}
	length_proj = 0;
	for(k=0;k<no_proj;k++)
	{
		if(k==0)
		{
			proj[k].p = 0;
			proj[k].q = 1;
		}
		else
		{
			proj[k].p = 1;
			proj[k].q = pow(2,k-1);
		}
		proj[k].max = 0;
		proj[k].min = 0;
		// assume a regular rectangular image 
		for(i=0;i<Q;i+=Q-1)
			for(j=0;j<P;j+=P-1)
			{
				bin = -i*proj[k].p + j*proj[k].q;
				if(bin>proj[k].max)
					proj[k].max = bin;
				if(bin<proj[k].min)
					proj[k].min = bin;	
			}
		proj[k].length = proj[k].max - proj[k].min + 1;
		if(length_proj<proj[k].length)
			length_proj = proj[k].length;
	}

// sort projections for inversion later
	qsort((void *)proj,no_proj,sizeof(projPARAM),CompareSlope);

// allocate memory for projection
	if( (projection=(int *)calloc(no_proj*length_proj,sizeof(int))) == NULL )
	{
		printf("ERROR: Projection memory could not be allocated.\n");
		exit(0);
	}

// allocate memory for projection
	if( (projection_copy=(int *)calloc(no_proj*length_proj,sizeof(int))) == NULL )
	{
		printf("ERROR: Projection Copy memory could not be allocated.\n");
		exit(0);
	}
	
	for(k=0;k<no_proj;k++)
		proj[k].ORIGINptr = projection + k*length_proj + abs(proj[k].min);

//allocate memory for bin_pointer
	if( (bin_pointer=(int **)calloc(Q*no_proj,sizeof(int *))) == NULL )
	{
		printf("ERROR: Bin pointer could not be allocated.\n");
		exit(0);
	}
				
	// read in image
	ReadPGM(argv[3],image,P*Q);
	
	
t0 = CPUTIME;

// project image
	for(k=0;k<no_proj;k++)
	{
		tempPRJ = projection + k*length_proj;
		for(i=0;i<proj[k].length;i++)
			*tempPRJ++ = 0;
		for(i=proj[k].length;i<length_proj;i++)
			*tempPRJ++ = -1;
		
		tempPRJ = proj[k].ORIGINptr;
		tempIMG = image;
		for(i=0;i<Q;i++)
		{
			for(j=0;j<P;j++)
			{
				*tempPRJ += *tempIMG++;
				*tempPRJ = (*tempPRJ+256*256)%256;
				tempPRJ += proj[k].q;
			}
			tempPRJ -= proj[k].p + P*proj[k].q;
		}
	}
t1 = CPUTIME;
time(&u2);
printf("CPU time to complete transform = %0.6f sec.\n",t1-t0,(int)(u2-u1));

// add noise
//	for(k=0;k<no_proj;k++)
//		for(i=0;i<proj[k].length;i++)
//			projection[k*length_proj + i] += GaussNoiseInt(0,3);

/* -------------------------------------------------------------------------------*/

Q_n = 2*Q;

ImageCopy(projection_copy,projection,no_proj*length_proj);

for(no_proj=log2(Q)+1;no_proj>=2;no_proj--)
{	
	Q_n /= 2;

	for(k=0;k<no_proj;k++)
	{
		if(k==0)
		{
			proj[k].p = 0;
			proj[k].q = 1;
		}
		else
		{
			proj[k].p = 1;
			proj[k].q = pow(2,k-1);
		}
		proj[k].max = 0;
		proj[k].min = 0;
		// assume a regular rectangular image 
		for(i=0;i<Q_n;i+=Q_n-1)
			for(j=0;j<P;j+=P-1)
			{
				bin = -i*proj[k].p + j*proj[k].q;
				if(bin>proj[k].max)
					proj[k].max = bin;
				if(bin<proj[k].min)
					proj[k].min = bin;	
			}
		proj[k].length = proj[k].max - proj[k].min + 1;
	}

// sort projections for inversion later
	qsort((void *)proj,no_proj,sizeof(projPARAM),CompareSlope);
	
	for(k=0;k<no_proj;k++)
		proj[k].ORIGINptr = projection + k*length_proj + abs(proj[k].min);

// re-initialise image data
	for(i=0;i<Q_n;i++)
		for(j=0;j<P;j++)
			image[i*P+j] = 0;

 // print projections
	j = pow(2,log2(Q)+1-no_proj);
	for(k=1;k<no_proj;k++)
	{
		sprintf(filename,"./output/proj[%d,%d]-level-%d.pgm",proj[k].p,j*proj[k].q,no_proj-1);
		WriteScaledPGM(filename,projection_copy+k*length_proj,j*proj[k].q,(proj[k].length+proj[k].q-1)/proj[k].q);
	}

t0 = CPUTIME;

// generate dependancy list
	proj[0].row_low = 0;
	proj[0].row_high = proj[0].q-1;
	for(i=1;i<no_proj;i++)
	{
		proj[i].row_low = proj[i-1].row_high + 1;
		proj[i].row_high = proj[i-1].row_high + proj[i].q;
	}

	i=0;
	for(l=0;l<Q_n;l++)
	{
		if(l<=proj[i].row_high)
			Depend[l].proj_no = i;
		else
		{
			i++;
			Depend[l].proj_no = i;
		}
		Depend[l].done = 0;
	}

// find rendezvous rows
	r_min = 0;
	while(((double)proj[r_min+1].p/(double)proj[r_min+1].q)<=0.0 && r_min<no_proj-1)
		r_min++;
	if(((double)proj[r_min].p/(double)proj[r_min].q)==0.0)
		r_max = r_min;
	else
	{
		r_max = no_proj - 1;
		while(((double)proj[r_max-1].p/(double)proj[r_max-1].q)>=0.0 && r_max>0)
			r_max--;
	}

// generate negative path

	l = proj[r_min].row_high;
	Depend[l].x = 0;
	
	for(i=r_min;i>=0;i--)
	{
		y = l;
		x = Depend[l].x;
		for(j=0;j<proj[i].q;j++)
		{
			Depend[l-j].x = (int)ceil((double)x - (double)(j*proj[i].p)/(double)proj[i].q);
		}
		if(l >= proj[i].q)
		{
			l -= proj[i].q;
			Depend[l].x = x - proj[i].p;
		}
	}

// generate positive path

	l = proj[r_max].row_low;
	Depend[l].x = 0;
	
	for(i=r_max;i<no_proj;i++)
	{
		y = l;
		x = Depend[l].x;
		for(j=0;j<proj[i].q;j++)
		{
			Depend[l+j].x = (int)ceil((double)x + (double)(j*proj[i].p)/(double)proj[i].q);
		}
		if(l < Q_n-proj[i].q)
		{
			l += proj[i].q;
			Depend[l].x = x + proj[i].p;
		}
	}

// determine offset
	offset = 0;
	for(i=0;i<Q_n;i++)
		if(Depend[i].x>offset)
			offset = Depend[i].x;
// shift reconstruction path back by offset
	for(i=0;i<Q_n;i++)
		Depend[i].x -= offset;

// determine reconstruction path of convex data
	k=0;
	while(k<Q_n)
	{
		for(i=0;i<Q_n;i++)
		{
			if(Depend[i].done==0)
			{
				l=1;
				x = Depend[i].x-proj[Depend[i].proj_no].p;
				y = i-proj[Depend[i].proj_no].q;
				while(y>=0)
				{
					if(Depend[y].x<=x)
						l = 0;
					x -= proj[Depend[i].proj_no].p;
					y -= proj[Depend[i].proj_no].q;			
				}
				x = Depend[i].x+proj[Depend[i].proj_no].p;
				y = i+proj[Depend[i].proj_no].q;
				while(y<Q_n)
				{
					if(Depend[y].x<=x)
						l = 0;
					x += proj[Depend[i].proj_no].p;
					y += proj[Depend[i].proj_no].q;
				}
				if(l==1) //point can be reconstructed
				{
					Recon[k].x = Depend[i].x;
					Recon[k].y = i;
					Recon[k].proj_no = Depend[i].proj_no;
					Recon[k].PIXELptr = image + i*P;
					Depend[i].x++;
					Depend[i].done = 1;
					k++;
				}
			}
		}
	}

	RECONptr = Recon;
	BINptr = bin_pointer;
	for(l=0;l<Q_n;l++)
	{
		PROJptr = proj;
		for(j=0;j<no_proj;j++)
		{
			BINptr[0] = PROJptr[0].ORIGINptr - RECONptr[0].y*PROJptr[0].p;
			BINptr++;
			PROJptr++;
		}
		RECONptr++;
	}

	for(l=0;l<P+offset;l++)
	{
		RECONptr = Recon;
		BINptr = bin_pointer;
		for(k=0;k<Q_n;k++)
		{
			if(RECONptr[0].x>=0 && RECONptr[0].x<P)
			{
				*RECONptr[0].PIXELptr = *BINptr[RECONptr[0].proj_no];
				PROJptr = proj;
				for(m=0;m<no_proj;m++)
				{
					*BINptr[0] -= *RECONptr[0].PIXELptr;
					*BINptr[0] = (*BINptr[0]+256*256)%256;
					BINptr[0] += PROJptr[0].q;
					BINptr++;
					PROJptr++;
				}
				RECONptr[0].PIXELptr++;
			}
			else
				BINptr+=no_proj;
				
			RECONptr[0].x++;
			RECONptr++;
		}

	}
t1 = CPUTIME;
time(&u2);
printf("CPU time to complete reconstruction = %0.6f sec.\n",t1-t0,(int)(u2-u1));

//expand in y-direction
	for(i=Q_n-1;i>=0;i--)
		for(k=0;k<Q/Q_n;k++)
		{
			for(j=0;j<P;j++)
			{
				image[(i*(Q/Q_n) + k)*P+j] = (image[i*P+j])%256;	
			}
		}
sprintf(filename,"./output/Reconstruction-level-%d.pgm",no_proj-1);
WriteScaledPGM(filename,image,P,Q);

	// low level data for next projection

	for(j=0;j<no_proj-1;j++)
	{
		temp_row = proj[j].q*pow(2,log2(Q)+1-no_proj);
		temp_col = (proj[j].length+proj[j].q-1)/proj[j].q;
		if(j==0)
			ImageCopy(projection,projection_copy,length_proj);
		else
		{
			for(i=0;i<temp_col;i++)
			{
				for(k=0;k<proj[j].q/2;k++)
					projection[j*length_proj+i*temp_row+k] = projection_copy[j*length_proj+i*temp_row+2*k] + projection_copy[j*length_proj+i*temp_row+2*k+1];
				for(k=proj[j].q/2;k<proj[j].q;k++)
					projection[j*length_proj+i*temp_row+k] = projection_copy[j*length_proj+i*temp_row+2*k] - projection_copy[j*length_proj+i*temp_row+2*k+1];
				for(k=proj[j].q;k<temp_row;k++)
					projection[j*length_proj+i*temp_row+k] = projection_copy[j*length_proj+i*temp_row+k];

			}
//			for(i=proj[j].length;i<length_proj;i++)
//				projection[j*length_proj+i] = 0;
		}
	}
	ImageCopy(projection_copy,projection,no_proj*length_proj);
	for(j=0;j<no_proj-1;j++)
	{
		temp_row = proj[j].q*pow(2,log2(Q)+1-no_proj);
		temp_col = (proj[j].length+proj[j].q-1)/proj[j].q;
		if(j==0)
			ImageCopy(projection,projection_copy,length_proj);
		else
		{
			for(i=0;i<temp_col;i++)
			{
				for(k=0;k<proj[j].q/2;k++)
					projection[j*length_proj+i*proj[j].q/2+k] = (projection_copy[j*length_proj+i*temp_row+k])%256;
			}
		}
	}
}	

/* --------------------------------------------------------------------------------- */

	free(image);
	free(projection);
	free(projection_copy);
	free(proj);
	free(Recon);
	free(Depend);
	free(bin_pointer);
	
	return 1;
}

