// performs nicolas balayage algorithm (general case)

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <time.h>

#include "../library/nrutil.h"
#include "../library/prototypes.h"
#include "../library/2D(k,th)transforms.h"
#include "../library/ProjInfo.h"
#include "../library/PGM.h"


#define CPUTIME (getrusage(RUSAGE_SELF,&ruse), ruse.ru_utime.tv_sec + ruse.ru_stime.tv_sec + 1e-6 * (ruse.ru_utime.tv_usec + ruse.ru_stime.tv_usec))

struct rusage ruse;

extern int getrusage();

typedef struct
{
	int p;
	int q;
	int length;
	int max;
	int min;
	int height;
	int *ORIGINptr;
	int offset;
	int row_low;
	int row_high;
}projPARAM;

typedef struct
{
	int x;
	int y;
	int proj_no;
	int *PIXELptr;
}reconLIST;

typedef struct
{
	int x;
	int proj_no;
	int done;
}dependLIST;

typedef struct
{
	int symbol;
	int freq;
	double prob;
} entropy_data;

int GaussNoiseInt(int mean, int range)
{
	double x=1.0, y=1.0;
	while(y > exp(-24.5*x*x))
	{
		x = (double)rand()/(double)RAND_MAX - 0.5;
		y = (double)rand()/(double)RAND_MAX;
	}
	return floor(x*(double)range + 0.5) + mean;
}

void ImageTranspose(int *image,int Q,int P)
{
	int i, j, *buffer, *tempIN, *tempOUT;

	buffer = (int *)calloc(P*Q,sizeof(int));
	ImageCopy(buffer,image,P*Q);
	if(Q>=P)
	{
		tempIN = buffer;
		for(i=0;i<P;i++)
		{	
			tempOUT = image + i;
			for(j=0;j<Q;j++)
			{
				*tempOUT = *tempIN++;
				tempOUT += P;
			}
		}
	}
	else
	{
		tempOUT = image;
		for(i=0;i<Q;i++)
		{	
			tempIN = buffer + i;
			for(j=0;j<P;j++)
			{
				*tempOUT++ = *tempIN;
				tempIN += Q;
			}
		}
	}
	free(buffer);
}

int MinINT(int elem1, int elem2)
{
	if(elem1<=elem2)
		return elem1;
	else
		return elem2;
}

int MaxINT(int elem1, int elem2)
{
	if(elem1>=elem2)
		return elem1;
	else
		return elem2;
}

int SignINT(int elem1)
{
	if(elem1>=0)
		return 1;
	else
		return -1;
}

int CompareP(elem1,elem2)
projPARAM *elem1; projPARAM *elem2;
{
	int a=elem1[0].p,b=elem2[0].p;
	return (a>b) ? 1 : (a<b) ? -1 : 0;
}

int CompareSlope(elem1,elem2)
projPARAM *elem1; projPARAM *elem2;
{
	double a=(double)elem1[0].q/(double)elem1[0].p,b=(double)elem2[0].q/(double)elem2[0].p;
	return (a>b) ? 1 : (a<b) ? -1 : 0;
}

void WavSumDiffBasePrime(int *data, int prime, int P, int Q, int row_length)
{
	int i, j, k, l, update, Htrue, Hbuff, Ltrue, Lbuff, Dbuff, *buff_ptr, *data_copy;
	
	Ltrue = (P+prime-1)/prime;
	Htrue = P-Ltrue;
	Dbuff = Ltrue*prime;
	Lbuff = Ltrue;
	Hbuff = Dbuff-Lbuff;

	int *buffer = (int *)calloc(2*Dbuff,sizeof(int));
	data_copy = buffer+Dbuff;
	buff_ptr = buffer+Lbuff;
		
	for(j=0;j<Q;j++)
	{
		ImageCopy(data_copy,data+j*row_length,P);
		for(i=P;i<Dbuff;i++)
			data_copy[i]=data_copy[i-1];

//		get prime-even data
		for(i=0;i<Ltrue;i++)
		{
			buffer[i] = 0;
			for(k=0;k<prime;k++)
				buffer[i] += data_copy[prime*i+k];
		}
		
//		subtract average from prime-odd data
		for(l=0;l<Htrue;l++)
		{
			i = l/(prime-1);
			k = l%(prime-1)+1;
			buff_ptr[l] = data_copy[prime*i+k] - (int)floor((1.0*buffer[i])/(1.0*prime) + 0.5);
		}
		
//		add back to data
		for(i=0;i<Ltrue;i++)
			data[j*row_length+i] = buffer[i];
		for(i=0;i<Htrue;i++)
			data[j*row_length+Ltrue+i] = buff_ptr[i];
	}
	free(buffer);
}

void InvWavSumDiffBasePrime(int *data, int prime, int P, int Q, int row_length)
{
	int i, j, k, l, update, Htrue, Hbuff, Ltrue, Lbuff, Dbuff, *buff_ptr, *data_copy;
	
	Ltrue = (P+prime-1)/prime;
	Htrue = P-Ltrue;
	Dbuff = Ltrue*prime;
	Lbuff = Ltrue;
	Hbuff = Dbuff-Lbuff;

	int *buffer = (int *)calloc(2*Dbuff,sizeof(int));
	data_copy = buffer+Dbuff;
	buff_ptr = buffer+Lbuff;
	
	for(j=0;j<Q;j++)
	{
//		get prime-even data
		ImageCopy(buffer,data+j*row_length,Ltrue);
//		get prime-odd
		ImageCopy(buff_ptr,data+j*row_length+Ltrue,Htrue);
		if(Htrue%(prime-1)==0)
			for(i=Htrue;i<Hbuff;i++)
				buff_ptr[i] = 0;
		else
			for(i=Htrue;i<Hbuff;i++)
				buff_ptr[i] = buff_ptr[i-1];
						
//		predict prime-odd
		for(i=0;i<Hbuff;i++)
		{
			k = i%(prime-1)+1;
			l = i/(prime-1);
			buff_ptr[i] += (int)floor((1.0*buffer[l])/(1.0*prime) + 0.5);	
		}

//		add back to data
		for(i=0;i<Ltrue;i++)
		{
			for(k=1;k<prime;k++)
			{
				if(prime*i+k<P)
					data[j*row_length+prime*i+k] = buff_ptr[i*(prime-1)+k-1];
				buffer[i] -= buff_ptr[i*(prime-1)+k-1];
			}
			data[j*row_length+prime*i] = buffer[i];
		}
	}
	free(buffer);
}

void WavSumDiff2DPrime(int *data, int prime_x, int prime_y, int P, int Q, int row_length, int col_height)
{
	// transform rows
	WavSumDiffBasePrime(data,prime_x,P,Q,row_length);
	ImageTranspose(data,row_length,col_height);
	// transform cols
	WavSumDiffBasePrime(data,prime_y,Q,P,col_height);
	ImageTranspose(data,col_height,row_length);
}

void InvWavSumDiff2DPrime(int *data, int prime_x, int prime_y, int P, int Q, int row_length, int col_height)
{
	// inv transform cols
	ImageTranspose(data,row_length,col_height);
	InvWavSumDiffBasePrime(data,prime_y,Q,P,col_height);
	// inv transform rows
	ImageTranspose(data,col_height,row_length);
	InvWavSumDiffBasePrime(data,prime_x,P,Q,row_length);
}

int CompareData(elem1,elem2)
int *elem1; int *elem2;
{
	return (elem1[0]>elem2[0]) ? 1 : (elem1[0]<elem2[0]) ? -1 : 0;
}

double Entropy(int *input_data, int N)
{
// input an array of integer data and the number of data points: N
	int *temp, *data, min, max, i, j, number_symbols;
	entropy_data *histogram, *temp_hist;
	double sum_prob=0.0, entropy=0.0;
	FILE *filepointer;
	char filename[100];

	// allocate memory for data copy
	if( (data = (int *)calloc(N,sizeof(int))) == NULL )
	{
		printf("ERROR: Data copy memory could not be allocated.\n");
		exit(0);
	}
	ImageCopy(data,input_data,N);

	//sort data
	qsort((void *)data,N,sizeof(int),CompareData);
		
	min = data[0];
	max = data[N-1];

	//determine number of symbols
	temp = data + 1;
	number_symbols = 1;
	for(i=1;i<N;i++)
	{
		if(temp[0]>temp[-1])
			number_symbols++;
		temp++;
	}
	
	// allocate memory for pseudo-histogram
	if( (histogram = (entropy_data *)calloc(number_symbols,sizeof(entropy_data))) == NULL )
	{
		printf("ERROR: Histogram memory could not be allocated.\n");
		exit(0);
	}

	// determine data frequencies
	histogram[0].symbol = data[0];
	histogram[0].freq = 1;
	temp_hist = histogram;
	temp = data+1;
	for(i=1;i<N;i++)
	{
		if(temp[0]>temp[-1])
		{
			temp_hist++;
			temp_hist[0].symbol = temp[0];			
			temp_hist[0].freq = 1;
		}
		else
			temp_hist[0].freq++;
		temp++;
	}
	// determine probabilities (also check sum of prob = 1)
	for(i=0;i<number_symbols;i++)
	{
		histogram[i].prob = (double)histogram[i].freq/(double)N;
		sum_prob += histogram[i].prob;
		if(histogram[i].prob != 0.0)
			entropy += histogram[i].prob*log2(1.0/histogram[i].prob);
	}
	
	printf("\n Number of data points: %d \n Data range: %d to %d \n Number of symbols: %d \n Sum of probabilities: %f \n Entropy (in b.p.p.): %f \n 2 ^ ent: %f \n",N,min,max,number_symbols,sum_prob,entropy,pow(2,entropy));
	
	free(histogram);
	free(data);
	return entropy;
}



int main(int argc, void *argv[])
{
	int P, Q, Q_n, h, i, j, k, l, x, y, m, n, bin, no_proj, length_proj, temp, temp_row, temp_col, tempW, tempH, x_temp,y_temp,b_x,b_y;
	int *image, *projection, *projection_copy, *tempIMG, *tempPRJ, check, q_check;
	int **bin_pointer, **BINptr;
	int S_minus=0, S_plus=0, S_total, r, max_row, min_row, offset, start_of_Next, end_of_Next, NoGood, r_max, r_min, q_up_max, q_down_max;
	projPARAM *proj, *PROJptr;
	reconLIST *Recon, *RECONptr;
	dependLIST *Depend;
	double t0, t1, ent_bpp;
	time_t u1,u2;
	char filename[50];

// read in image width P
	if( sscanf(argv[1],"%d",&P) == 0 )
	{
		printf("ERROR: could not read in image width, P.\n");
		exit(0);
	}

// read in image height Q
	if( sscanf(argv[2],"%d",&Q) == 0 )
	{
		printf("ERROR: could not read in image height, Q.\n");
		exit(0);
	}

// allocate memory for image
	if( (image=(int *)calloc(P*Q,sizeof(int))) == NULL )
	{
		printf("ERROR: Image memory could not be allocated.\n");
		exit(0);
	}
	
// allocate memory for LISTS
	if( (Recon=(reconLIST *)calloc(Q,sizeof(reconLIST))) == NULL )
	{
		printf("ERROR: LIST memory could not be allocated.\n");
		exit(0);
	}
	if( (Depend=(dependLIST *)calloc(P*Q,sizeof(dependLIST))) == NULL )
	{
		printf("ERROR: LIST memory could not be allocated.\n");
		exit(0);
	}

// proj data

	// allocate memory for projection parameters
	if( (proj=(projPARAM *)calloc(no_proj,sizeof(projPARAM))) == NULL )
	{
		printf("ERROR: Projection parameter memory could not be allocated.\n");
		exit(0);
	}
	length_proj = 0;

	no_proj = 2;

	proj[0].p = 256;
	proj[0].q = 4;
	proj[1].p = -256;
	proj[1].q = 4;

//	no_proj = 4;
//
//	proj[0].p = 64;
//	proj[0].q = 64;
//	proj[1].p = -64;
//	proj[1].q = 64;
//	proj[2].p = 64;
//	proj[2].q = 128;
//	proj[3].p = -64;
//	proj[3].q = 128;
	
	for(k=0;k<no_proj;k++)
	{
		proj[k].max = 0;
		proj[k].min = 0;
		for(i=0;i<Q;i++)
			for(j=0;j<P;j++)
			{
				//assuming |p| > q
				x = j%abs(proj[k].p);
				y = i - proj[k].q*(j-x)/proj[k].p;
				//otherwise
				//	y = i%abs(proj[k].q);
				//	x = j - proj[k].p*(i-y)/proj[k].q;
				bin = x + y*abs(proj[k].p);
				if(bin>proj[k].max)
					proj[k].max = bin;
				if(bin<proj[k].min)
					proj[k].min = bin;
			}
		proj[k].length = proj[k].max - proj[k].min;
		proj[k].height = (proj[k].length+abs(proj[k].p)-1)/abs(proj[k].p);
		if(length_proj<proj[k].height*abs(proj[k].p))
			length_proj = proj[k].height*abs(proj[k].p);
		printf("%d, %d with min = %d, max = %d\n",proj[k].p,proj[k].q,proj[k].min,proj[k].max);
	}
	
// sort projections for inversion later
	qsort((void *)proj,no_proj,sizeof(projPARAM),CompareSlope);

// allocate memory for projection
	if( (projection=(int *)calloc(no_proj*length_proj,sizeof(int))) == NULL )
	{
		printf("ERROR: Projection memory could not be allocated.\n");
		exit(0);
	}

// allocate memory for projection
	if( (projection_copy=(int *)calloc(no_proj*length_proj,sizeof(int))) == NULL )
	{
		printf("ERROR: Projection Copy memory could not be allocated.\n");
		exit(0);
	}

//allocate memory for bin_pointer
	if( (bin_pointer=(int **)calloc(Q*no_proj,sizeof(int *))) == NULL )
	{
		printf("ERROR: Bin pointer could not be allocated.\n");
		exit(0);
	}
				
	// read in image
	ReadPGM(argv[3],image,P*Q);
	
//	tempW=0;
//	for(i=0;i<Q;i++)
//		for(j=0;j<P;j++)
//			tempW += image[i*P+j];
//
//	tempH = (int)floor((1.0*tempW)/(1.0*P*Q)+0.5);
//	printf("average value is: %d\n",tempH);
//	for(i=0;i<Q;i++)
//		for(j=0;j<P;j++)
//			image[i*P+j] -= tempH;
	
t0 = CPUTIME;

// project image
	for(k=0;k<no_proj;k++)
	{
		tempPRJ = projection + k*length_proj;
		for(i=0;i<proj[k].length;i++)
			*tempPRJ++ = 0;
		for(i=proj[k].length;i<length_proj;i++)
			*tempPRJ++ = 0;
		
		proj[k].ORIGINptr = projection + k*length_proj - proj[k].min;
		tempPRJ = proj[k].ORIGINptr;
		tempIMG = image;
		for(i=0;i<Q;i++)
		{
			tempPRJ = proj[k].ORIGINptr + i*abs(proj[k].p);
			for(j=0;j<P;j++)
			{
				*tempPRJ += *tempIMG++;
				tempPRJ++;
				if((j+1)%abs(proj[k].p)==0)
					tempPRJ -= proj[k].p*(proj[k].q) + abs(proj[k].p);
				//printf("pos [%2d]: %d\n",k,tempPRJ-proj[k].ORIGINptr);
			}
		}
		sprintf(filename,"./output/proj(%d,%d,b)-S.pgm",proj[k].p,proj[k].q);
		WriteScaledPGM(filename,projection + k*length_proj,abs(proj[k].p),proj[k].height);
		sprintf(filename,"./output/proj(%d,%d,b).pgm",proj[k].p,proj[k].q);
		WritePGM(filename,projection + k*length_proj,abs(proj[k].p),proj[k].height);
		printf("%d, %d with min = %d, max = %d\n",proj[k].p,proj[k].q,proj[k].min,proj[k].max);
	}
	
t1 = CPUTIME;
time(&u2);
printf("CPU time to complete transform = %0.6f sec.\n",t1-t0,(int)(u2-u1));

// entropy
	j = 0;
	for(k=0;k<no_proj;k++)
	{
		ImageCopy(projection_copy+j,projection+k*length_proj,proj[k].height*abs(proj[k].p));
		j+=proj[k].height*abs(proj[k].p);
	}
	ent_bpp = Entropy(projection_copy,j);
	printf("raw projections have %d bins with ent. of %5.2f giving compression ratio of %5.2f, or (%5.2f bpp)\n",j,ent_bpp,(100.0*j*ent_bpp)/(P*Q*8.0),(j*ent_bpp)/(1.0*P*Q));

// reconstruct (assuming projections ordered by gradient /--\

//	clear image
//	for(y=0;y<Q;y++)
//		for(x=0;x<P;x++)
//			image[y*P+x] = 0;
//	
//	first fill in convex hull
//	check = 0;
//	for(k=0;k<no_proj;k++)
//		if(SignINT(proj[k].p)==-1)
//			check++;
//	
//	for(k=0;k<check;k++)
//	{
//		for(y=0;y<proj[k].q;y++)
//			for(x=0;x<abs(proj[k].p);x++)
//			{
//				x_temp = k*abs(proj[k].p) + x;
//				y_temp = y;
//				if(y_temp <= Q/2)
//				{
//					temp = projection[k*length_proj+(y + k*proj[k].q)*abs(proj[k].p)+x];
//					image[y_temp*P+x_temp] = temp;
//					for(j=0;j<no_proj;j++)
//					{
//						b_x = x_temp%abs(proj[j].p);
//						b_y = y_temp - proj[j].q*(x_temp-b_x)/proj[j].p;
//						proj[j].ORIGINptr[b_y*abs(proj[j].p)+b_x] -= temp;
//					}
//				}
//				x_temp = (no_proj-1-k)*abs(proj[k].p) + x;
//				y_temp = Q-1-y;
//				if(y_temp > Q/2)
//				{
//					temp = projection[k*length_proj+(y_temp + (no_proj-1-k)*proj[k].q)*abs(proj[k].p)+x];
//					image[y_temp*P+x_temp] = temp;
//					for(j=0;j<no_proj;j++)
//					{
//						b_x = x_temp%abs(proj[j].p);
//						b_y = y_temp - proj[j].q*(x_temp-b_x)/proj[j].p;
//						proj[j].ORIGINptr[b_y*abs(proj[j].p)+b_x] -= temp;
//					}
//				}
//			}
//	}
//	for(k=no_proj-1;k>=check;k--)
//	{
//		for(y=0;y<proj[k].q;y++)
//			for(x=0;x<abs(proj[k].p);x++)
//			{
//				x_temp = k*abs(proj[k].p) + x;
//				y_temp = y;
//				if(y_temp <= Q/2)
//				{
//					temp = projection[k*length_proj+(y + (no_proj-1-k)*proj[k].q)*abs(proj[k].p)+x];
//					image[y_temp*P+x_temp] = temp;
//					for(j=0;j<no_proj;j++)
//					{
//						b_x = x_temp%abs(proj[j].p);
//						b_y = y_temp - proj[j].q*(x_temp-b_x)/proj[j].p;
//						proj[j].ORIGINptr[b_y*abs(proj[j].p)+b_x] -= temp;
//					}
//				}
//				x_temp = (no_proj-1-k)*abs(proj[k].p) + x;
//				y_temp = Q-1-y;
//				if(y_temp > Q/2)
//				{
//					temp = projection[k*length_proj+(y_temp + k*proj[k].q)*abs(proj[k].p)+x];
//					image[y_temp*P+x_temp] = temp;
//					for(j=0;j<no_proj;j++)
//					{
//						b_x = x_temp%abs(proj[j].p);
//						b_y = y_temp - proj[j].q*(x_temp-b_x)/proj[j].p;
//						proj[j].ORIGINptr[b_y*abs(proj[j].p)+b_x] -= temp;
//					}
//				}
//			}
//	}
//	
//	if(check<no_proj)
//	{
//		q_check = proj[check].q;
//		if(check>0 && proj[check-1].q<q_check)
//			q_check = proj[check-1].q;
//	}
//	else
//		q_check = proj[no_proj-1].q;
//		
//		
//	for(i=0;i<Q/(2*q_check);i++)
//	{
//		for(k=0;k<check;k++)
//		{
//			for(y=0;y<q_check;y++)
//				for(x=0;x<abs(proj[k].p);x++)
//				{
//					x_temp = k*abs(proj[k].p) + x;
//					y_temp = y+proj[k].q+i*q_check;
//					if(y_temp <= Q/2)
//					{
//						temp = projection[k*length_proj+(y_temp + k*proj[k].q)*abs(proj[k].p)+x];
//						image[y_temp*P+x_temp] = temp;
//						for(j=0;j<no_proj;j++)
//						{
//							b_x = x_temp%abs(proj[j].p);
//							b_y = y_temp - proj[j].q*(x_temp-b_x)/proj[j].p;
//							proj[j].ORIGINptr[b_y*abs(proj[j].p)+b_x] -= temp;
//						}
//					}
//					x_temp = (no_proj-1-k)*abs(proj[k].p) + x;
//					y_temp = Q-1-y-proj[k].q-i*q_check;
//					if(y_temp > Q/2)
//					{
//						temp = projection[k*length_proj+(y_temp + (no_proj-1-k)*proj[k].q)*abs(proj[k].p)+x];
//						image[y_temp*P+x_temp] = temp;
//						for(j=0;j<no_proj;j++)
//						{
//							b_x = x_temp%abs(proj[j].p);
//							b_y = y_temp - proj[j].q*(x_temp-b_x)/proj[j].p;
//							proj[j].ORIGINptr[b_y*abs(proj[j].p)+b_x] -= temp;
//						}
//					}
//				}
//		}
//		for(k=no_proj-1;k>=check;k--)
//		{
//			for(y=0;y<q_check;y++)
//				for(x=0;x<abs(proj[k].p);x++)
//				{
//					x_temp = k*abs(proj[k].p) + x;
//					y_temp = y+proj[k].q+i*q_check;
//					if(y_temp <= Q/2)
//					{
//						temp = projection[k*length_proj+(y_temp + (no_proj-1-k)*proj[k].q)*abs(proj[k].p)+x];
//						image[y_temp*P+x_temp] = temp;
//						for(j=0;j<no_proj;j++)
//						{
//							b_x = x_temp%abs(proj[j].p);
//							b_y = y_temp - proj[j].q*(x_temp-b_x)/proj[j].p;
//							proj[j].ORIGINptr[b_y*abs(proj[j].p)+b_x] -= temp;
//						}
//					}
//					x_temp = (no_proj-1-k)*abs(proj[k].p) + x;
//					y_temp = Q-1-y-proj[k].q-i*q_check;
//					if(y_temp > Q/2)
//					{
//						temp = projection[k*length_proj+(y_temp + k*proj[k].q)*abs(proj[k].p)+x];
//						image[y_temp*P+x_temp] = temp;
//						for(j=0;j<no_proj;j++)
//						{
//							b_x = x_temp%abs(proj[j].p);
//							b_y = y_temp - proj[j].q*(x_temp-b_x)/proj[j].p;
//							proj[j].ORIGINptr[b_y*abs(proj[j].p)+b_x] -= temp;
//						}
//					}
//				}
//		}
//	}
//	sprintf(filename,"./output/image-R.pgm");
//	WritePGM(filename,image,P,Q);

// wavelet transform projections

	l = 2;
	for(k=0;k<no_proj;k++)
	{
		tempW = abs(proj[k].p);
		tempH = proj[k].height;
		
		for(i=0;i<l;i++)
		{
			if(i%2==1)
			{
				WavSumDiff2DPrime(projection+k*length_proj,2,2,tempW,tempH,abs(proj[k].p),proj[k].height);
				tempW = (tempW+1)/2;
				tempH = (tempH+1)/2;
			}
			else
			{
				WavSumDiff2DPrime(projection+k*length_proj,2,2,tempW,tempH,abs(proj[k].p),proj[k].height);
				tempW = (tempW+1)/2;
				tempH = (tempH+1)/2;
			}
		}
		sprintf(filename,"./output/proj(%d,%d,b)-W-S.pgm",proj[k].p,proj[k].q);
		WriteScaledPGM(filename,projection + k*length_proj,abs(proj[k].p),proj[k].height);
		//WritePGM(filename,projection + k*length_proj,abs(proj[k].p),proj[k].height);
		sprintf(filename,"./output/proj(%d,%d,b)-W-O.pgm",proj[k].p,proj[k].q);
		WriteOffsetPGM(filename,projection + k*length_proj,abs(proj[k].p),proj[k].height,128);
	}

// entropy
	j = 0;
	for(k=0;k<no_proj;k++)
	{
		ImageCopy(projection_copy+j,projection+k*length_proj,proj[k].height*abs(proj[k].p));
		j+=proj[k].height*abs(proj[k].p);
	}
	ent_bpp = Entropy(projection_copy,j);
	printf("DWT projections have %d bins with ent. of %5.2f giving compression ratio of %5.2f, or (%5.2f bpp)\n",j,ent_bpp,(100.0*j*ent_bpp)/(P*Q*8.0),(j*ent_bpp)/(1.0*P*Q));

//	//copy lower resolution projections to projection_copy
//	for(k=0;k<no_proj;k++)
//	{
//		proj[k].length = (proj[k].length+3)/4;
//		proj[k].height = (proj[k].height+1)/2;
//		for(x=0;x<abs(proj[k].p)/2;x++)
//			for(y=0;y<proj[k].height;y++)
//				projection_copy[k*length_proj+y*(abs(proj[k].p)/2)+x] = (int)floor(1.0*projection[k*length_proj+y*abs(proj[k].p)+x]/4.0 + 0.5);
//		proj[k].p /= 2;
//		proj[k].q /= 2;
//		proj[k].ORIGINptr = projection_copy + k*length_proj - proj[k].min/4;
//		sprintf(filename,"./output/proj(%d,%d,b)-L-S.pgm",proj[k].p,proj[k].q);
//		WriteScaledPGM(filename,projection_copy + k*length_proj,abs(proj[k].p),proj[k].height);
//	}
//
//	P = (P+1)/2;
//	Q = (Q+1)/2;

	// put zeros into low-res projection to projection_copy
	for(k=0;k<no_proj;k++)
	{
		for(x=0;x<abs(proj[k].p)/2;x++)
			for(y=0;y<proj[k].height/2;y++)
				projection[k*length_proj+y*abs(proj[k].p)+x] = 0;
		sprintf(filename,"./output/proj(%d,%d,b)-ZERO-S.pgm",proj[k].p,proj[k].q);
		WriteScaledPGM(filename,projection + k*length_proj,abs(proj[k].p),proj[k].height);
	}

	l = 1;
	for(k=0;k<no_proj;k++)
	{
		tempW = abs(proj[k].p);
		tempH = proj[k].height;
				
		for(i=0;i<l;i++)
		{
			if(i%2==1)
			{
				InvWavSumDiff2DPrime(projection+k*length_proj,2,2,tempW,tempH,abs(proj[k].p),proj[k].height);
				tempW *= 2;
				tempH *= 2;
			}
			else
			{
				InvWavSumDiff2DPrime(projection+k*length_proj,2,2,tempW,tempH,abs(proj[k].p),proj[k].height);
				tempW *= 2;
				tempH *= 2;
			}
		}
		sprintf(filename,"./output/proj(%d,%d,b)-ZERO-W-S.pgm",proj[k].p,proj[k].q);
		WriteScaledPGM(filename,projection + k*length_proj,abs(proj[k].p),proj[k].height);
		//WritePGM(filename,projection + k*length_proj,abs(proj[k].p),proj[k].height);
		sprintf(filename,"./output/proj(%d,%d,b)-ZERO-W-O.pgm",proj[k].p,proj[k].q);
		WriteOffsetPGM(filename,projection + k*length_proj,abs(proj[k].p),proj[k].height,128);
	}

	for(k=0;k<no_proj;k++)
	{
		ImageCopy(projection_copy+k*length_proj,projection+k*length_proj,proj[k].height*abs(proj[k].p));
	}

//	clear image
	for(y=0;y<Q;y++)
		for(x=0;x<P;x++)
			image[y*P+x] = 0;
	
//	first fill in convex hull
	check = 0;
	for(k=0;k<no_proj;k++)
	{
		proj[k].ORIGINptr += projection_copy - projection;
		if(SignINT(proj[k].p)==-1)
			check++;
	}
	for(k=0;k<check;k++)
	{
		for(y=0;y<proj[k].q;y++)
			for(x=0;x<abs(proj[k].p);x++)
			{
				x_temp = k*abs(proj[k].p) + x;
				y_temp = y;
				if(y_temp < Q)
				{
					temp = projection_copy[k*length_proj+(y + k*proj[k].q)*abs(proj[k].p)+x];
					image[y_temp*P+x_temp] = temp;
					for(j=0;j<no_proj;j++)
					{
						b_x = x_temp%abs(proj[j].p);
						b_y = y_temp - proj[j].q*(x_temp-b_x)/proj[j].p;
						proj[j].ORIGINptr[b_y*abs(proj[j].p)+b_x] -= temp;
					}
				}
			}
	}
	for(k=no_proj-1;k>=check;k--)
	{
		for(y=0;y<proj[k].q;y++)
			for(x=0;x<abs(proj[k].p);x++)
			{
				x_temp = k*abs(proj[k].p) + x;
				y_temp = y;
				if(y_temp < Q)
				{
					temp = projection_copy[k*length_proj+(y + (no_proj-1-k)*proj[k].q)*abs(proj[k].p)+x];
					image[y_temp*P+x_temp] = temp;
					for(j=0;j<no_proj;j++)
					{
						b_x = x_temp%abs(proj[j].p);
						b_y = y_temp - proj[j].q*(x_temp-b_x)/proj[j].p;
						proj[j].ORIGINptr[b_y*abs(proj[j].p)+b_x] -= temp;
					}
				}
			}
	}

//	then fill in the rest
	
	if(check<no_proj)
	{
		q_check = proj[check].q;
		if(check>0 && proj[check-1].q<q_check)
			q_check = proj[check-1].q;
	}
	else
		q_check = proj[no_proj-1].q;
		
	for(i=0;i<Q/q_check;i++)
	{
		for(k=0;k<check;k++)
		{
			for(y=0;y<q_check;y++)
				for(x=0;x<abs(proj[k].p);x++)
				{
					x_temp = k*abs(proj[k].p) + x;
					y_temp = y+proj[k].q+i*q_check;
					if(y_temp < Q)
					{
						temp = projection_copy[k*length_proj+(y_temp + k*proj[k].q)*abs(proj[k].p)+x];
						image[y_temp*P+x_temp] = temp;
						for(j=0;j<no_proj;j++)
						{
							b_x = x_temp%abs(proj[j].p);
							b_y = y_temp - proj[j].q*(x_temp-b_x)/proj[j].p;
							proj[j].ORIGINptr[b_y*abs(proj[j].p)+b_x] -= temp;
						}
					}
				}
		}
		for(k=no_proj-1;k>=check;k--)
		{
			for(y=0;y<q_check;y++)
				for(x=0;x<abs(proj[k].p);x++)
				{
					x_temp = k*abs(proj[k].p) + x;
					y_temp = y+proj[k].q+i*q_check;
					if(y_temp < Q)
					{
						temp = projection_copy[k*length_proj+(y_temp + (no_proj-1-k)*proj[k].q)*abs(proj[k].p)+x];
						image[y_temp*P+x_temp] = temp;
						for(j=0;j<no_proj;j++)
						{
							b_x = x_temp%abs(proj[j].p);
							b_y = y_temp - proj[j].q*(x_temp-b_x)/proj[j].p;
							proj[j].ORIGINptr[b_y*abs(proj[j].p)+b_x] -= temp;
						}
					}
				}
		}
	}
	sprintf(filename,"./output/image-R-S.pgm");
	WriteScaledPGM(filename,image,P,Q);

/* -------------------------------------------------------------------------------*/

	free(image);
	free(projection);
	free(projection_copy);
	free(proj);
	free(Recon);
	free(Depend);
	free(bin_pointer);
	
	return 1;
}

