// performs nicolas balayage algorithm (general case)

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <time.h>

#include "../library/nrutil.h"
#include "../library/prototypes.h"
#include "../library/2D(k,th)transforms.h"
#include "../library/ProjInfo.h"
#include "../library/PGM.h"


#define CPUTIME (getrusage(RUSAGE_SELF,&ruse), ruse.ru_utime.tv_sec + ruse.ru_stime.tv_sec + 1e-6 * (ruse.ru_utime.tv_usec + ruse.ru_stime.tv_usec))

struct rusage ruse;

extern int getrusage();

typedef struct
{
	int p;
	int q;
	int length;
	int max;
	int min;
	int height;
	int *ORIGINptr;
	int offset;
	int row_low;
	int row_high;
}projPARAM;

typedef struct
{
	int x;
	int y;
	int proj_no;
	int *PIXELptr;
}reconLIST;

typedef struct
{
	int x;
	int proj_no;
	int done;
}dependLIST;

typedef struct
{
	int symbol;
	int freq;
	double prob;
} entropy_data;

int GaussNoiseInt(int mean, int range)
{
	double x=1.0, y=1.0;
	while(y > exp(-24.5*x*x))
	{
		x = (double)rand()/(double)RAND_MAX - 0.5;
		y = (double)rand()/(double)RAND_MAX;
	}
	return floor(x*(double)range + 0.5) + mean;
}

void ImageTranspose(int *image,int Q,int P)
{
	int i, j, *buffer, *tempIN, *tempOUT;

	buffer = (int *)calloc(P*Q,sizeof(int));
	ImageCopy(buffer,image,P*Q);
	if(Q>=P)
	{
		tempIN = buffer;
		for(i=0;i<P;i++)
		{	
			tempOUT = image + i;
			for(j=0;j<Q;j++)
			{
				*tempOUT = *tempIN++;
				tempOUT += P;
			}
		}
	}
	else
	{
		tempOUT = image;
		for(i=0;i<Q;i++)
		{	
			tempIN = buffer + i;
			for(j=0;j<P;j++)
			{
				*tempOUT++ = *tempIN;
				tempIN += Q;
			}
		}
	}
	free(buffer);
}

int MinINT(int elem1, int elem2)
{
	if(elem1<=elem2)
		return elem1;
	else
		return elem2;
}

int MaxINT(int elem1, int elem2)
{
	if(elem1>=elem2)
		return elem1;
	else
		return elem2;
}

int CompareP(elem1,elem2)
projPARAM *elem1; projPARAM *elem2;
{
	int a=elem1[0].p,b=elem2[0].p;
	return (a>b) ? 1 : (a<b) ? -1 : 0;
}

int CompareSlope(elem1,elem2)
projPARAM *elem1; projPARAM *elem2;
{
	double a=(double)elem1[0].p/(double)elem1[0].q,b=(double)elem2[0].p/(double)elem2[0].q;
	return (a>b) ? 1 : (a<b) ? -1 : 0;
}

void LiftWavBasePrime(int *data, int prime, int P, int Q, int row_length)
{
	int i, j, k, l, update, Htrue, Hbuff, Ltrue, Lbuff, Dbuff, *buff_ptr, *data_copy;
	int *buffer = (int *)calloc(2*(P+2*prime),sizeof(int));
	
	Ltrue = (P+prime-1)/prime;
	Htrue = P-Ltrue;
	Dbuff = ((P-1+prime-1)/prime)*prime+1;
	Lbuff = (Dbuff+prime-1)/prime;
	Hbuff = Dbuff-Lbuff;
	
	data_copy = buffer+P+2*prime;
	buff_ptr = buffer+Lbuff;
	
	for(j=0;j<Q;j++)
	{
		ImageCopy(data_copy,data+j*row_length,P);

//		get prime-even data
		for(i=0;i<Ltrue;i++)
			buffer[i] = data_copy[prime*i];
		for(i=Ltrue;i<Lbuff;i++)
			buffer[i] = buffer[i-1];
		
//		predict prime-odd
		for(l=0;l<Htrue;l++)
		{
			i = l/(prime-1);
			k = l%(prime-1)+1;
			buff_ptr[l] = data_copy[prime*i+k]-((prime-k)*buffer[i]+k*buffer[i+1])/prime;
		}
		for(l=Htrue;l<Hbuff;l++)
			buff_ptr[l] = 0;
		
//		update prime-even
//		NOTE: 1 + 2 + 3 + ... + prime-1 = prime(prime-1)/2;
//		first
		update = 0;
		for(k=1;k<prime;k++)
			update += k*buff_ptr[k-1];
		buffer[0] += update/prime;
//		rest
		for(l=1;l<Ltrue;l++)
		{
			update = 0;
			for(k=1;k<prime;k++)
				update += k*(buff_ptr[(prime-1)*l+(k-1)] + buff_ptr[(prime-1)*l-k]);
			buffer[l] += update/(2*prime);
		}

//		add back to data
		for(i=0;i<Ltrue;i++)
			data[j*row_length+i] = buffer[i];
		for(i=0;i<Htrue;i++)
			data[j*row_length+Ltrue+i] = buff_ptr[i] + 128;
	}
	free(buffer);
}

void InvLiftWavBasePrime(int *data, int prime, int P, int Q, int row_length)
{
	int i, j, k, l, update, Htrue, Hbuff, Ltrue, Lbuff, Dbuff, *buff_ptr, *data_copy;
	int *buffer = (int *)calloc(2*(P+2*prime),sizeof(int));
	
	Ltrue = (P+prime-1)/prime;
	Htrue = P-Ltrue;
	Dbuff = ((P-1+prime-1)/prime)*prime+1;
	Lbuff = (Dbuff+prime-1)/prime;
	Hbuff = Dbuff-Lbuff;
	
	data_copy = buffer+P+2*prime;
	buff_ptr = buffer+Lbuff;
	
	for(j=0;j<Q;j++)
	{
//		get prime-even data
		ImageCopy(buffer,data+j*row_length,Ltrue);
//		get prime-odd
		ImageCopy(buff_ptr,data+j*row_length+Ltrue,Htrue);
		for(i=0;i<Htrue;i++)
			buff_ptr[i] -= 128;
		for(i=Htrue;i<Hbuff;i++)
			buff_ptr[i] = 0;
			
//		update prime-even
//		NOTE: 1 + 2 + 3 + ... + prime-1 = prime(prime-1)/2;
//		first
		update = 0;
		for(k=1;k<prime;k++)
			update += k*buff_ptr[k-1];
		buffer[0] -= update/prime;
//		rest
		for(l=1;l<Ltrue;l++)
		{
			update = 0;
			for(k=1;k<prime;k++)
				update += k*(buff_ptr[(prime-1)*l+(k-1)] + buff_ptr[(prime-1)*l-k]);
			buffer[l] -= update/(2*prime);
		}
		for(l=Ltrue;l<Lbuff;l++)
			buffer[l] = buffer[l-1];			
			
//		predict prime-odd
		for(i=0;i<Htrue;i++)
		{
			k = i%(prime-1)+1;
			l = i/(prime-1);
			buff_ptr[i] += ((prime-k)*buffer[l]+k*buffer[l+1])/prime;	
		}

//		add back to data
		for(i=0;i<Ltrue;i++)
		{
			data[j*row_length+prime*i] = buffer[i];
			for(k=1;k<prime;k++)
				if(prime*i+k<P)
					data[j*row_length+prime*i+k] = buff_ptr[i*(prime-1)+k-1];
		}
	}
	free(buffer);
}

void LiftWav2DPrime(int *data, int prime_x, int prime_y, int P, int Q, int row_length, int col_height)
{
	// transform rows
	LiftWavBasePrime(data,prime_x,P,Q,row_length);
	ImageTranspose(data,row_length,col_height);
	// transform cols
	LiftWavBasePrime(data,prime_y,Q,P,col_height);
	ImageTranspose(data,col_height,row_length);
}


void InvLiftWav2DPrime(int *data, int prime_x, int prime_y, int P, int Q, int row_length, int col_height)
{
	// inv transform cols
	ImageTranspose(data,row_length,col_height);
	InvLiftWavBasePrime(data,prime_y,Q,P,col_height);
	// inv transform rows
	ImageTranspose(data,col_height,row_length);
	InvLiftWavBasePrime(data,prime_x,P,Q,row_length);
}

int CompareData(elem1,elem2)
int *elem1; int *elem2;
{
	return (elem1[0]>elem2[0]) ? 1 : (elem1[0]<elem2[0]) ? -1 : 0;
}

double Entropy(int *input_data, int N)
{
// input an array of integer data and the number of data points: N
	int *temp, *data, min, max, i, j, number_symbols;
	entropy_data *histogram, *temp_hist;
	double sum_prob=0.0, entropy=0.0;
	FILE *filepointer;
	char filename[100];

	// allocate memory for data copy
	if( (data = (int *)calloc(N,sizeof(int))) == NULL )
	{
		printf("ERROR: Data copy memory could not be allocated.\n");
		exit(0);
	}
	ImageCopy(data,input_data,N);

	//sort data
	qsort((void *)data,N,sizeof(int),CompareData);
		
	min = data[0];
	max = data[N-1];

	//determine number of symbols
	temp = data + 1;
	number_symbols = 1;
	for(i=1;i<N;i++)
	{
		if(temp[0]>temp[-1])
			number_symbols++;
		temp++;
	}
	
	// allocate memory for pseudo-histogram
	if( (histogram = (entropy_data *)calloc(number_symbols,sizeof(entropy_data))) == NULL )
	{
		printf("ERROR: Histogram memory could not be allocated.\n");
		exit(0);
	}

	// determine data frequencies
	histogram[0].symbol = data[0];
	histogram[0].freq = 1;
	temp_hist = histogram;
	temp = data+1;
	for(i=1;i<N;i++)
	{
		if(temp[0]>temp[-1])
		{
			temp_hist++;
			temp_hist[0].symbol = temp[0];			
			temp_hist[0].freq = 1;
		}
		else
			temp_hist[0].freq++;
		temp++;
	}
	// determine probabilities (also check sum of prob = 1)
	for(i=0;i<number_symbols;i++)
	{
		histogram[i].prob = (double)histogram[i].freq/(double)N;
		sum_prob += histogram[i].prob;
		if(histogram[i].prob != 0.0)
			entropy += histogram[i].prob*log2(1.0/histogram[i].prob);
	}
	
	printf("\n Number of data points: %d \n Data range: %d to %d \n Number of symbols: %d \n Sum of probabilities: %f \n Entropy (in b.p.p.): %f \n 2 ^ ent: %f \n",N,min,max,number_symbols,sum_prob,entropy,pow(2,entropy));
	
	free(histogram);
	free(data);
	return entropy;
}



int main(int argc, void *argv[])
{
	int P, Q, Q_n, h, i, j, k, l, x, y, m, n, bin, no_proj, length_proj, temp, temp_row, temp_col, tempW, tempH;
	int *image, *projection, *projection_copy, *tempIMG, *tempPRJ;
	int **bin_pointer, **BINptr;
	int S_minus=0, S_plus=0, S_total, r, max_row, min_row, offset, start_of_Next, end_of_Next, NoGood, r_max, r_min, q_up_max, q_down_max;
	projPARAM *proj, *PROJptr;
	reconLIST *Recon, *RECONptr;
	dependLIST *Depend;
	double t0, t1, ent_bpp;
	time_t u1,u2;
	char filename[50];

// read in image width P
	if( sscanf(argv[1],"%d",&P) == 0 )
	{
		printf("ERROR: could not read in image width, P.\n");
		exit(0);
	}

// read in image height Q
	if( sscanf(argv[2],"%d",&Q) == 0 )
	{
		printf("ERROR: could not read in image height, Q.\n");
		exit(0);
	}

// allocate memory for image
	if( (image=(int *)calloc(P*Q,sizeof(int))) == NULL )
	{
		printf("ERROR: Image memory could not be allocated.\n");
		exit(0);
	}
	
// allocate memory for LISTS
	if( (Recon=(reconLIST *)calloc(Q,sizeof(reconLIST))) == NULL )
	{
		printf("ERROR: LIST memory could not be allocated.\n");
		exit(0);
	}
	if( (Depend=(dependLIST *)calloc(P*Q,sizeof(dependLIST))) == NULL )
	{
		printf("ERROR: LIST memory could not be allocated.\n");
		exit(0);
	}

// read in proj data
	no_proj = log2(Q)+1;

	// allocate memory for projection parameters
	if( (proj=(projPARAM *)calloc(no_proj,sizeof(projPARAM))) == NULL )
	{
		printf("ERROR: Projection parameter memory could not be allocated.\n");
		exit(0);
	}
	length_proj = 0;
	
	proj[0].p = 1;
	proj[0].q = 0;
	proj[0].min = 0;
	proj[0].max = Q;
	proj[0].length = Q;
	proj[0].height = Q;

	proj[1].p = 1;
	proj[1].q = 1;
	proj[0].min = -(P+1)/2;
	proj[0].max = Q;
	proj[0].length = proj[0].max - proj[0].min;
	proj[0].height = proj[0].max - proj[0].min;
	
	l = 1;
	for(k=2;k<no_proj;k++)
	{
		if((k+no_proj)%2==1)
			l*= 3;

		proj[k].p = proj[k-1].p*2;
		proj[k].q = l;
		
		proj[k].max = Q*proj[k].p;
		i = proj[k].p-1;
		proj[k].min = 0;
		while(i+proj[k].p<P)
		{
			proj[k].min -= proj[k].q*proj[k].p;
			i += proj[k].p;
		}
		// assume a regular rectangular image  
		proj[k].length = proj[k].max - proj[k].min;
		proj[k].height = proj[k].length/proj[k].p;	
		if(length_proj<proj[k].length)
			length_proj = proj[k].length;
		printf("%d, %d with min = %d, max = %d\n",proj[k].p,proj[k].q,proj[k].min,proj[k].max);
	}

// sort projections for inversion later
	qsort((void *)proj,no_proj,sizeof(projPARAM),CompareSlope);

// allocate memory for projection
	if( (projection=(int *)calloc(no_proj*length_proj,sizeof(int))) == NULL )
	{
		printf("ERROR: Projection memory could not be allocated.\n");
		exit(0);
	}

// allocate memory for projection
	if( (projection_copy=(int *)calloc(no_proj*length_proj,sizeof(int))) == NULL )
	{
		printf("ERROR: Projection Copy memory could not be allocated.\n");
		exit(0);
	}
	
	for(k=0;k<no_proj;k++)
		proj[k].ORIGINptr = projection + k*length_proj + MaxINT(0,-proj[k].min);

//allocate memory for bin_pointer
	if( (bin_pointer=(int **)calloc(Q*no_proj,sizeof(int *))) == NULL )
	{
		printf("ERROR: Bin pointer could not be allocated.\n");
		exit(0);
	}
				
	// read in image
	ReadPGM(argv[3],image,P*Q);
	
	
t0 = CPUTIME;

// project image
	for(k=0;k<no_proj;k++)
	{
		tempPRJ = projection + k*length_proj;
		for(i=0;i<proj[k].length;i++)
			*tempPRJ++ = 0;
		for(i=proj[k].length;i<length_proj;i++)
			*tempPRJ++ = -1;
		
		tempPRJ = proj[k].ORIGINptr;
		tempIMG = image;
		for(i=0;i<Q;i++)
		{
			tempPRJ = proj[k].ORIGINptr + i*proj[k].p;
			for(j=0;j<P;j++)
			{
				*tempPRJ += *tempIMG++;
				tempPRJ++;
				if((j+1)%proj[k].p==0)
					tempPRJ -= proj[k].p*(proj[k].q+1);
			}
		}
		sprintf(filename,"./output/proj(%d,%d,b)-S.pgm",proj[k].p,proj[k].q);
		WriteScaledPGM(filename,projection + k*length_proj,proj[k].p,proj[k].height);
		printf("%d, %d with min = %d, max = %d\n",proj[k].p,proj[k].q,proj[k].min,proj[k].max);
	}
	
t1 = CPUTIME;
time(&u2);
printf("CPU time to complete transform = %0.6f sec.\n",t1-t0,(int)(u2-u1));

// entropy
	j = 0;
	for(k=0;k<no_proj;k++)
	{
		ImageCopy(projection_copy+j,projection+k*length_proj,proj[k].length);
		j+=proj[k].length;
	}
	ent_bpp = Entropy(projection_copy,j);
	printf("raw projections have %d bins with ent. of %5.2f giving compression ratio of %5.2f, or (%5.2f bpp)\n",j,ent_bpp,(100.0*j*ent_bpp)/(P*Q*8.0),(j*ent_bpp)/(1.0*P*Q));


// wavelet transform projections

// first proj
	for(i=0;i<proj[0].length-1;i++)
		projection[i] -= projection[i+1];

// other projs
	for(k=1;k<no_proj-1;k++)
	{
		tempW = proj[k].p;
		tempH = proj[k].height;
		if(k==1)
			l = 2;
		else if(k==no_proj-1)
			l = no_proj-2;
		else if(k%2==0)
			l = k+2;
		else
			l = k-2;
			 
		for(i=no_proj;i>no_proj-l;i--)
		{
			if(i%2==1)
			{
				LiftWav2DPrime(projection+k*length_proj,2,1,tempW,tempH,proj[k].p,proj[k].height);
				tempW = (tempW+1)/2;
			}
			else
			{
				LiftWav2DPrime(projection+k*length_proj,2,3,tempW,tempH,proj[k].p,proj[k].height);
				tempW = (tempW+1)/2;
				tempH = (tempH+2)/3;
			}
		}
		sprintf(filename,"./output/proj(%d,%d,b)-W.pgm",proj[k].p,proj[k].q);
		WritePGM(filename,projection + k*length_proj,proj[k].p,proj[k].height);
	}

// last proj
	tempW = proj[no_proj-1].height;
	for(i=no_proj;i>no_proj-l;i--)
	{
		LiftWavBasePrime(projection+(no_proj-1)*length_proj,2,tempW,1,proj[k].height);
		tempW = (tempW+1)/2;
	}

// entropy
	j = 0;
	for(k=0;k<no_proj;k++)
	{
		ImageCopy(projection_copy+j,projection+k*length_proj,proj[k].length);
		j+=proj[k].length;
	}
	ent_bpp = Entropy(projection_copy,j);
	printf("DWT projections have %d bins with ent. of %5.2f giving compression ratio of %5.2f, or (%5.2f bpp)\n",j,ent_bpp,(100.0*j*ent_bpp)/(P*Q*8.0),(j*ent_bpp)/(1.0*P*Q));


/* -------------------------------------------------------------------------------*/

	free(image);
	free(projection);
	free(projection_copy);
	free(proj);
	free(Recon);
	free(Depend);
	free(bin_pointer);
	
	return 1;
}

