// performs nicolas balayage algorithm (general case)

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <time.h>

#include "../library/nrutil.h"
#include "../library/prototypes.h"
#include "../library/2D(k,th)transforms.h"
#include "../library/ProjInfo.h"
#include "../library/PGM.h"


#define CPUTIME (getrusage(RUSAGE_SELF,&ruse), ruse.ru_utime.tv_sec + ruse.ru_stime.tv_sec + 1e-6 * (ruse.ru_utime.tv_usec + ruse.ru_stime.tv_usec))

struct rusage ruse;

extern int getrusage();

typedef struct
{
	int p;
	int q;
	int length;
	int max;
	int min;
	int *ORIGINptr;
	int offset;
	int row_low;
	int row_high;
}projPARAM;

typedef struct
{
	int x;
	int y;
	int proj_no;
	int *PIXELptr;
}reconLIST;

typedef struct
{
	int x;
	int proj_no;
	int done;
}dependLIST;


typedef struct
{
	int symbol;
	int freq;
	double prob;
} entropy_data;


int CompareData(elem1,elem2)
int *elem1; int *elem2;
{
	return (elem1[0]>elem2[0]) ? 1 : (elem1[0]<elem2[0]) ? -1 : 0;
}

double Entropy(int *input_data, int N)
{
// input an array of integer data and the number of data points: N
	int *temp, *data, min, max, i, j, number_symbols;
	entropy_data *histogram, *temp_hist;
	double sum_prob=0.0, entropy=0.0;
	FILE *filepointer;
	char filename[100];

	// allocate memory for data copy
	if( (data = (int *)calloc(N,sizeof(int))) == NULL )
	{
		printf("ERROR: Data copy memory could not be allocated.\n");
		exit(0);
	}
	ImageCopy(data,input_data,N);

	//sort data
	qsort((void *)data,N,sizeof(int),CompareData);
		
	min = data[0];
	max = data[N-1];

	//determine number of symbols
	temp = data + 1;
	number_symbols = 1;
	for(i=1;i<N;i++)
	{
		if(temp[0]>temp[-1])
			number_symbols++;
		temp++;
	}
	
	// allocate memory for pseudo-histogram
	if( (histogram = (entropy_data *)calloc(number_symbols,sizeof(entropy_data))) == NULL )
	{
		printf("ERROR: Histogram memory could not be allocated.\n");
		exit(0);
	}

	// determine data frequencies
	histogram[0].symbol = data[0];
	histogram[0].freq = 1;
	temp_hist = histogram;
	temp = data+1;
	for(i=1;i<N;i++)
	{
		if(temp[0]>temp[-1])
		{
			temp_hist++;
			temp_hist[0].symbol = temp[0];			
			temp_hist[0].freq = 1;
		}
		else
			temp_hist[0].freq++;
		temp++;
	}
	// determine probabilities (also check sum of prob = 1)
	for(i=0;i<number_symbols;i++)
	{
		histogram[i].prob = (double)histogram[i].freq/(double)N;
		sum_prob += histogram[i].prob;
		if(histogram[i].prob != 0.0)
			entropy += histogram[i].prob*log2(1.0/histogram[i].prob);
	}
	
	printf("\n Number of data points: %d \n Data range: %d to %d \n Number of symbols: %d \n Sum of probabilities: %f \n Entropy (in b.p.p.): %f \n 2 ^ ent: %f \n",N,min,max,number_symbols,sum_prob,entropy,pow(2,entropy));
	
	free(histogram);
	free(data);
	return entropy;
}


int GaussNoiseInt(int mean, int range)
{
	double x=1.0, y=1.0;
	while(y > exp(-24.5*x*x))
	{
		x = (double)rand()/(double)RAND_MAX - 0.5;
		y = (double)rand()/(double)RAND_MAX;
	}
	return floor(x*(double)range + 0.5) + mean;
}

void ImageTranspose(int *image,int Q,int P)
{
	int i, j, *buffer, *tempIN, *tempOUT;

	buffer = (int *)calloc(P*Q,sizeof(int));
	ImageCopy(buffer,image,P*Q);
	if(Q>=P)
	{
		tempIN = buffer;
		for(i=0;i<P;i++)
		{	
			tempOUT = image + i;
			for(j=0;j<Q;j++)
			{
				*tempOUT = *tempIN++;
				tempOUT += P;
			}
		}
	}
	else
	{
		tempOUT = image;
		for(i=0;i<Q;i++)
		{	
			tempIN = buffer + i;
			for(j=0;j<P;j++)
			{
				*tempOUT++ = *tempIN;
				tempIN += Q;
			}
		}
	}
	free(buffer);
}


void LiftWavBase2(int *data, int P, int Q)
{
	int i, j, Htrue, Hbuff, Ltrue, Lbuff, Dbuff, *buff_ptr, *data_copy;
	int *buffer = (int *)calloc(2*(P+3),sizeof(int));
	
	Ltrue = (P+1)/2;
	Htrue = P-Ltrue;
	Dbuff = 2*Ltrue+1;
	Lbuff = Dbuff/2;
	Hbuff = Dbuff-Lbuff-1;
	
	data_copy = buffer+P+3;
	buff_ptr = buffer+Lbuff;
	
	for(j=0;j<Q;j++)
	{
	
		ImageCopy(data_copy,data+j*P,P);
		for(i=P;i<Dbuff;i++)
			data_copy[i] = data_copy[i-2];
	
		//get even data
		for(i=0;i<Lbuff;i++)
			buffer[i] = data_copy[2*i];
		
		//predict odd
		for(i=0;i<Hbuff;i++)
			buff_ptr[i] =  data_copy[2*i+1]-(buffer[i]+buffer[i+1])/2;	

		//update even
		//first
			buffer[0] += buff_ptr[0]/2;
		//rest
		for(i=1;i<Lbuff;i++)
			buffer[i] += (buff_ptr[i-1]+buff_ptr[i])/4;
		
		//add back to data
		for(i=0;i<Ltrue;i++)
			data[j*P+i] = buffer[i];
		for(i=0;i<Htrue;i++)
			data[j*P+Ltrue+i] = 128+buff_ptr[i];
	}
	free(buffer);
}

void LiftWavBasePrime(int *data, int prime, int P, int Q, int row_length)
{
	int i, j, k, l, update, Htrue, Hbuff, Ltrue, Lbuff, Dbuff, *buff_ptr, *data_copy;
	int *buffer = (int *)calloc(2*(P+2*prime),sizeof(int));
	
	Ltrue = (P+prime-1)/prime;
	Htrue = P-Ltrue;
	Dbuff = ((P-1+prime-1)/prime)*prime+1;
	Lbuff = (Dbuff+prime-1)/prime;
	Hbuff = Dbuff-Lbuff;
	
	data_copy = buffer+P+2*prime;
	buff_ptr = buffer+Lbuff;
	
	for(j=0;j<Q;j++)
	{
		ImageCopy(data_copy,data+j*row_length,P);

//		get prime-even data
		for(i=0;i<Ltrue;i++)
			buffer[i] = data_copy[prime*i];
		for(i=Ltrue;i<Lbuff;i++)
			buffer[i] = buffer[i-1];
		
//		predict prime-odd
		for(l=0;l<Htrue;l++)
		{
			i = l/(prime-1);
			k = l%(prime-1)+1;
			buff_ptr[l] = data_copy[prime*i+k]-((prime-k)*buffer[i]+k*buffer[i+1])/prime;
		}
		for(l=Htrue;l<Hbuff;l++)
			buff_ptr[l] = 0;
		
//		update prime-even
//		NOTE: 1 + 2 + 3 + ... + prime-1 = prime(prime-1)/2;
//		first
		update = 0;
		for(k=1;k<prime;k++)
			update += k*buff_ptr[k-1];
		buffer[0] += update/prime;
//		rest
		for(l=1;l<Ltrue;l++)
		{
			update = 0;
			for(k=1;k<prime;k++)
				update += floor(exp(-(prime-k)/(0.36*prime))*(buff_ptr[(prime-1)*l+(k-1)] + buff_ptr[(prime-1)*l-k])/(0.48*prime) + 0.5);
			buffer[l] += update;
		}

//		add back to data
		for(i=0;i<Ltrue;i++)
			data[j*row_length+i] = buffer[i];
		for(i=0;i<Htrue;i++)
			data[j*row_length+Ltrue+i] = buff_ptr[i] + 128;
	}
	free(buffer);
}

void InvLiftWavBasePrime(int *data, int prime, int P, int Q, int row_length)
{
	int i, j, k, l, update, Htrue, Hbuff, Ltrue, Lbuff, Dbuff, *buff_ptr, *data_copy;
	int *buffer = (int *)calloc(2*(P+2*prime),sizeof(int));
	
	Ltrue = (P+prime-1)/prime;
	Htrue = P-Ltrue;
	Dbuff = ((P-1+prime-1)/prime)*prime+1;
	Lbuff = (Dbuff+prime-1)/prime;
	Hbuff = Dbuff-Lbuff;
	
	data_copy = buffer+P+2*prime;
	buff_ptr = buffer+Lbuff;
	
	for(j=0;j<Q;j++)
	{
//		get prime-even data
		ImageCopy(buffer,data+j*row_length,Ltrue);
//		get prime-odd
		ImageCopy(buff_ptr,data+j*row_length+Ltrue,Htrue);
		for(i=0;i<Htrue;i++)
			buff_ptr[i] -= 128;
		for(i=Htrue;i<Hbuff;i++)
			buff_ptr[i] = 0;
			
//		update prime-even
//		NOTE: 1 + 2 + 3 + ... + prime-1 = prime(prime-1)/2;
//		first
		update = 0;
		for(k=1;k<prime;k++)
			update += k*buff_ptr[k-1];
		buffer[0] -= update/prime;
//		rest
		for(l=1;l<Ltrue;l++)
		{
			update = 0;
			for(k=1;k<prime;k++)
				update += floor(exp(-(prime-k)/(0.36*prime))*(buff_ptr[(prime-1)*l+(k-1)] + buff_ptr[(prime-1)*l-k])/(0.48*prime) + 0.5);
			buffer[l] -= update;
		}
		for(l=Ltrue;l<Lbuff;l++)
			buffer[l] = buffer[l-1];			
			
//		predict prime-odd
		for(i=0;i<Htrue;i++)
		{
			k = i%(prime-1)+1;
			l = i/(prime-1);
			buff_ptr[i] += ((prime-k)*buffer[l]+k*buffer[l+1])/prime;	
		}

//		add back to data
		for(i=0;i<Ltrue;i++)
		{
			data[j*row_length+prime*i] = buffer[i];
			for(k=1;k<prime;k++)
				if(prime*i+k<P)
					data[j*row_length+prime*i+k] = buff_ptr[i*(prime-1)+k-1];
		}
	}
	free(buffer);
}

void LiftWav2DPrime(int *data, int prime_x, int prime_y, int P, int Q, int row_length, int col_height)
{
	// transform rows
	LiftWavBasePrime(data,prime_x,P,Q,row_length);
	ImageTranspose(data,row_length,col_height);
	// transform cols
	LiftWavBasePrime(data,prime_y,Q,P,col_height);
	ImageTranspose(data,col_height,row_length);
}

void InvLiftWav2DPrime(int *data, int prime_x, int prime_y, int P, int Q, int row_length, int col_height)
{
	// inv transform cols
	ImageTranspose(data,row_length,col_height);
	InvLiftWavBasePrime(data,prime_y,Q,P,col_height);
	// inv transform rows
	ImageTranspose(data,col_height,row_length);
	InvLiftWavBasePrime(data,prime_x,P,Q,row_length);
}



int main(int argc, void *argv[])
{
	int P, Q, i, j, tempW, tempH, no_loops;
	int *image;
		
// read in image width P
	if( sscanf(argv[1],"%d",&P) == 0 )
	{
		printf("ERROR: could not read in image width, P.\n");
		exit(0);
	}

// read in image height Q
	if( sscanf(argv[2],"%d",&Q) == 0 )
	{
		printf("ERROR: could not read in image height, Q.\n");
		exit(0);
	}

// allocate memory for image
	if( (image=(int *)calloc(P*Q,sizeof(int))) == NULL )
	{
		printf("ERROR: Image memory could not be allocated.\n");
		exit(0);
	}
	
	// read in image
	ReadPGM(argv[3],image,P*Q);
	Entropy(image,P*Q);
	WritePGM("./output/image.pgm",image,P,Q);
	
	tempW = P;
	tempH = Q;
	no_loops = 6;
	for(i=0;i<no_loops;i++)
	{
		printf("%2d: W = %4d, H = %4d\n",i,tempW,tempH);
		if(i%2==1)
		{
			LiftWav2DPrime(image,2,1,tempW,tempH,P,Q);
			tempW = (tempW+1)/2;
		}
		else
		{
			LiftWav2DPrime(image,2,3,tempW,tempH,P,Q);
			tempW = (tempW+1)/2;
			tempH = (tempH+2)/3;
		}
	}
	// output
	Entropy(image,P*Q);
	WritePGM("./output/WavLifted-S.pgm",image,P,Q);

	for(i=no_loops-1;i>=0;i--)
	{
		tempW = P;
		tempH = Q;
		for(j=0;j<i;j++)
			if(j%2==1)
			{
				tempW = (tempW+1)/2;
			}
			else
			{
				tempW = (tempW+1)/2;
				tempH = (tempH+2)/3;
			}
		printf("%2d: W = %4d, H = %4d\n",i,tempW,tempH);

		if(i%2==1)
		{
			InvLiftWav2DPrime(image,2,1,tempW,tempH,P,Q);
			tempW = (tempW+1)/2;
		}
		else
		{
			InvLiftWav2DPrime(image,2,3,tempW,tempH,P,Q);
			tempW = (tempW+1)/2;
			tempH = (tempH+2)/3;
		}
	}
	// output
	WritePGM("./output/InvWavLifted-S.pgm",image,P,Q);


//	j = 11;
//	for(i=0;i<7*j;i++)
//		image[i] = 0;
//	
//	for(i=-j+1;i<j;i++)
//	{
//		if(i==0)
//			image[3*j+i] += 1000;
//		else
//		{
//			image[3*j+i] += floor(1000.0*(exp(-abs(i)/(0.36*j)))/(0.48*j) + 0.5);
//			image[((3*j+i)/j)*j] -= floor(1000.0*(j-(j+i)%j)*(exp(-abs(i)/(0.36*j)))/(0.48*j*j) + 0.5);
//			image[((4*j+i)/j)*j] -= floor(1000.0*((j+i)%j)*(exp(-abs(i)/(0.36*j)))/(0.48*j*j) + 0.5);
//		}
//	}
//
//	P = 0;
//	for(i=0;i<7*j;i++)
//	{
//		printf("%4d\n",image[i]);
//		P += image[i];
//	}
//	printf("sum is %4d\n",P);


	free(image);
		
	return 123;
}

