// performs nicolas balayage algorithm (general case)

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <time.h>

#include "../library/nrutil.h"
#include "../library/prototypes.h"
#include "../library/2D(k,th)transforms.h"
#include "../library/ProjInfo.h"
#include "../library/PGM.h"


#define CPUTIME (getrusage(RUSAGE_SELF,&ruse), ruse.ru_utime.tv_sec + ruse.ru_stime.tv_sec + 1e-6 * (ruse.ru_utime.tv_usec + ruse.ru_stime.tv_usec))

struct rusage ruse;

extern int getrusage();

typedef struct
{
	int p;
	int q;
	int length;
	int max;
	int min;
	int *ORIGINptr;
	int offset;
	int row_low;
	int row_high;
}projPARAM;

typedef struct
{
	int x;
	int y;
	int proj_no;
	int *PIXELptr;
}reconLIST;

typedef struct
{
	int x;
	int proj_no;
	int done;
}dependLIST;


int GaussNoiseInt(int mean, int range)
{
	double x=1.0, y=1.0;
	while(y > exp(-24.5*x*x))
	{
		x = (double)rand()/(double)RAND_MAX - 0.5;
		y = (double)rand()/(double)RAND_MAX;
	}
	return floor(x*(double)range + 0.5) + mean;
}

void ImageTranspose(int *image,int Q,int P)
{
	int i, j, *buffer, *tempIN, *tempOUT;

	buffer = (int *)calloc(P*Q,sizeof(int));
	ImageCopy(buffer,image,P*Q);
	if(Q>=P)
	{
		tempIN = buffer;
		for(i=0;i<P;i++)
		{	
			tempOUT = image + i;
			for(j=0;j<Q;j++)
			{
				*tempOUT = *tempIN++;
				tempOUT += P;
			}
		}
	}
	else
	{
		tempOUT = image;
		for(i=0;i<Q;i++)
		{	
			tempIN = buffer + i;
			for(j=0;j<P;j++)
			{
				*tempOUT++ = *tempIN;
				tempIN += Q;
			}
		}
	}
	free(buffer);
}


void WavSumDiffBasePrime(int *data, int prime, int P, int Q, int row_length)
{
	int i, j, k, l, update, Htrue, Hbuff, Ltrue, Lbuff, Dbuff, *buff_ptr, *data_copy;
	
	Ltrue = (P+prime-1)/prime;
	Htrue = P-Ltrue;
	Dbuff = Ltrue*prime;
	Lbuff = Ltrue;
	Hbuff = Dbuff-Lbuff;

	int *buffer = (int *)calloc(2*Dbuff,sizeof(int));
	data_copy = buffer+Dbuff;
	buff_ptr = buffer+Lbuff;
		
	for(j=0;j<Q;j++)
	{
		ImageCopy(data_copy,data+j*row_length,P);
		for(i=P;i<Dbuff;i++)
			data_copy[i]=data_copy[i-1];

//		get prime-even data
		for(i=0;i<Ltrue;i++)
		{
			buffer[i] = 0;
			for(k=0;k<prime;k++)
				buffer[i] += data_copy[prime*i+k];
		}
		
//		subtract average from prime-odd data
		for(l=0;l<Htrue;l++)
		{
			i = l/(prime-1);
			k = l%(prime-1)+1;
			buff_ptr[l] = data_copy[prime*i+k] - (int)floor((1.0*buffer[i])/(1.0*prime) + 0.5);
		}
		
//		add back to data
		for(i=0;i<Ltrue;i++)
			data[j*row_length+i] = buffer[i];
		for(i=0;i<Htrue;i++)
			data[j*row_length+Ltrue+i] = buff_ptr[i];
	}
	free(buffer);
}

void InvWavSumDiffBasePrime(int *data, int prime, int P, int Q, int row_length)
{
	int i, j, k, l, update, Htrue, Hbuff, Ltrue, Lbuff, Dbuff, *buff_ptr, *data_copy;
	
	Ltrue = (P+prime-1)/prime;
	Htrue = P-Ltrue;
	Dbuff = Ltrue*prime;
	Lbuff = Ltrue;
	Hbuff = Dbuff-Lbuff;

	int *buffer = (int *)calloc(2*Dbuff,sizeof(int));
	data_copy = buffer+Dbuff;
	buff_ptr = buffer+Lbuff;
	
	for(j=0;j<Q;j++)
	{
//		get prime-even data
		ImageCopy(buffer,data+j*row_length,Ltrue);
//		get prime-odd
		ImageCopy(buff_ptr,data+j*row_length+Ltrue,Htrue);
		if(Htrue%(prime-1)==0)
			for(i=Htrue;i<Hbuff;i++)
				buff_ptr[i] = 0;
		else
			for(i=Htrue;i<Hbuff;i++)
				buff_ptr[i] = buff_ptr[i-1];
						
//		predict prime-odd
		for(i=0;i<Hbuff;i++)
		{
			k = i%(prime-1)+1;
			l = i/(prime-1);
			buff_ptr[i] += (int)floor((1.0*buffer[l])/(1.0*prime) + 0.5);	
		}

//		add back to data
		for(i=0;i<Ltrue;i++)
		{
			for(k=1;k<prime;k++)
			{
				if(prime*i+k<P)
					data[j*row_length+prime*i+k] = buff_ptr[i*(prime-1)+k-1];
				buffer[i] -= buff_ptr[i*(prime-1)+k-1];
			}
			data[j*row_length+prime*i] = buffer[i];
		}
	}
	free(buffer);
}

void WavSumDiff2DPrime(int *data, int prime_x, int prime_y, int P, int Q, int row_length, int col_height)
{
	// transform rows
	WavSumDiffBasePrime(data,prime_x,P,Q,row_length);
	ImageTranspose(data,row_length,col_height);
	// transform cols
	WavSumDiffBasePrime(data,prime_y,Q,P,col_height);
	ImageTranspose(data,col_height,row_length);
}

void InvWavSumDiff2DPrime(int *data, int prime_x, int prime_y, int P, int Q, int row_length, int col_height)
{
	// inv transform cols
	ImageTranspose(data,row_length,col_height);
	InvWavSumDiffBasePrime(data,prime_y,Q,P,col_height);
	// inv transform rows
	ImageTranspose(data,col_height,row_length);
	InvWavSumDiffBasePrime(data,prime_x,P,Q,row_length);
}

int main(int argc, void *argv[])
{
	int P, Q, i, j, tempW, tempH;
	int *image;
		
// read in image width P
	if( sscanf(argv[1],"%d",&P) == 0 )
	{
		printf("ERROR: could not read in image width, P.\n");
		exit(0);
	}

// read in image height Q
	if( sscanf(argv[2],"%d",&Q) == 0 )
	{
		printf("ERROR: could not read in image height, Q.\n");
		exit(0);
	}

// allocate memory for image
	if( (image=(int *)calloc(P*Q,sizeof(int))) == NULL )
	{
		printf("ERROR: Image memory could not be allocated.\n");
		exit(0);
	}
	
	// read in image
	ReadPGM(argv[3],image,P*Q);
	WritePGM("./output/image.pgm",image,P,Q);
	
	tempW=0;
	for(i=0;i<Q;i++)
		for(j=0;j<P;j++)
			tempW += image[i*P+j];

	tempH = (int)floor((1.0*tempW)/(1.0*P*Q)+0.5);
	printf("average value is: %d\n",tempH);
	for(i=0;i<Q;i++)
		for(j=0;j<P;j++)
			image[i*P+j] -= tempH;

			
	tempW = P;
	tempH = Q;
	for(i=0;i<9;i++)
	{
		printf("%2d: W = %4d, H = %4d\n",i,tempW,tempH);
		if(i%2==1)
		{
			WavSumDiff2DPrime(image,2,1,tempW,tempH,P,Q);
			tempW = (tempW+1)/2;
		}
		else
		{
			WavSumDiff2DPrime(image,2,3,tempW,tempH,P,Q);
			tempW = (tempW+1)/2;
			tempH = (tempH+2)/3;
		}
	}
	// output
	WriteScaledPGM("./output/WavLifted-S.pgm",image,P,Q);

	for(i=8;i>=0;i--)
	{
		tempW = P;
		tempH = Q;
		for(j=0;j<i;j++)
			if(j%2==1)
			{
				tempW = (tempW+1)/2;
			}
			else
			{
				tempW = (tempW+1)/2;
				tempH = (tempH+2)/3;
			}
		printf("%2d: W = %4d, H = %4d\n",i,tempW,tempH);

		if(i%2==1)
		{
			InvWavSumDiff2DPrime(image,2,1,tempW,tempH,P,Q);
			tempW = (tempW+1)/2;
		}
		else
		{
			InvWavSumDiff2DPrime(image,2,3,tempW,tempH,P,Q);
			tempW = (tempW+1)/2;
			tempH = (tempH+2)/3;
		}
	}
	// output
	WritePGM("./output/InvWavLifted-S.pgm",image,P,Q);

	free(image);
		
	return 123;
}

